""" handle parameters like transforming, merging, converting"""
from dataclasses import dataclass

from collections import OrderedDict
import mdpeditor.mdpblocks.provide


@dataclass
class CompiledParameters:

    """The result of compiling parameters,
       keeping track of meta during the compilation process like key duplications
    """
    parameters: OrderedDict
    duplicate_keys: set


def compile_parameters(option_blocks):

    """merge block by block into parameters, keeping track of duplicates

    Args:
        option_blocks (iterable over OrderedDict): Options to merge
    """

    parameters = OrderedDict()  # type: OrderedDict
    duplicate_keys = set()

    for block in option_blocks:

        # collect overlapping parameters in the blocks
        duplicate_keys |= set(parameters.keys()) & set(block.keys())
        parameters.update(block)

    return CompiledParameters(parameters, duplicate_keys)


class OutputParameters:
    """
    Handle parameter output like printing and cleaning
    """
    def __init__(self, parameters):

        # open default .mdp and fill in the "other parameters"
        # discard all non-default parameters
        self.parameters = mdpeditor.mdpblocks.provide.default_parameter_block()

        for key in self.parameters.keys():
            if key in parameters.keys():
                self.parameters[key] = parameters[key]

    def keep_only(self, keys_to_keep):
        """ Remove all keys that are not to be kept """
        keys_to_pop = [key for key in self.parameters.keys() if key not in keys_to_keep]
        for key in keys_to_pop:
            self.parameters.pop(key)

    def as_string(self):
        """ write the parameters as a string in .mdp format"""
        max_key_length = max(map(len, self.parameters.keys()))

        formatted_mdp_entries = [
            f'{key:{max_key_length+1}s} = {value}'
            for key, value in self.parameters.items()
        ]

        return '\n'.join(formatted_mdp_entries) + "\n"

    def as_colored_string(self, keys_to_color):
        """ write the parameters as a string in .mdp format"""
        max_key_length = max(map(len, self.parameters.keys()))
        color = "\u001b[32m"  # green
        reset = "\u001b[0m"

        formatted_mdp_entries = [
            f'{color}{key:{max_key_length+1}s} = {value}{reset}'
            if key in keys_to_color else f'{key:{max_key_length+1}s} = {value}'
            for key, value in self.parameters.items()
        ]

        return '\n'.join(formatted_mdp_entries) + "\n"



""" Allow to access the contents of pre-defined mdp parameter blocks"""
import importlib.resources
import configparser
import io
import sys

from collections import OrderedDict


def available_parameter_blocks():
    """
    List available parameter blocks in the package resource
    """

    available_mdp_blocks = []

    # the mdp blocks are stored as a sub-package with non-python data
    # in nested folders like default/parameters.mdp
    mdpblock_package_contents = importlib.resources.contents(__package__)

    for package_content in mdpblock_package_contents:

        # go through all subfolders
        if not importlib.resources.is_resource(__package__, package_content):
            # look for .mdp files
            for block in importlib.resources.contents(__package__ + "." +
                                                      package_content):
                if block[-4:] == ".mdp":
                    available_mdp_blocks += [
                        package_content + "." + block[:-4]
                    ]
    available_mdp_blocks.sort()
    return available_mdp_blocks


def default_parameter_block():
    """
    Return a parameter block that has all parameters set to default values
    """

    return mdp_string_to_ordered_dict(
        parameter_block_as_string("default.parameters"))


def read_parameter_blocks(blocks):
    """
    Read a list of specified parameter blocks as a mapping to an OrderedDict.
    """

    if not blocks:
        return map(mdp_string_to_ordered_dict, [""])

    mdp_option_blocks_as_strings = [
        parameter_block_as_string(block_to_compile)
        for block_to_compile in blocks
    ]

    return map(mdp_string_to_ordered_dict, mdp_option_blocks_as_strings)


def describe_parameter_blocks() -> str:
    """Return a string with a self-description of all parameter blocks
    """
    description_string = ""
    for block in available_parameter_blocks():
        if sys.stdout.isatty():
            # make things bold
            description_string += '\033[1m'

        description_string += f'{block}'

        if sys.stdout.isatty():
            # back to normal formatting
            description_string += '\033[0m'

        description_string += "\n\t"
        
        # print the first lines of the parameter file that are prepended with
        # a semicolon, omitting the semicolon
        block_as_string = parameter_block_as_string(block)
        for line in block_as_string.splitlines():
            if len(line) > 0 and (line[0] == ';'):
                description_string += line[1:]
            else:
                # newline for next block
                description_string += "\n\n"
                break

    return description_string


def parameter_block_as_string(block_name):
    """
    Return an .mdp parameter block from the package resources
    """

    return importlib.resources.read_text(
        __package__ + "." + block_name.split(".")[0],
        block_name.split(".")[1] + ".mdp")


def mdp_string_to_ordered_dict(mdp_string: str) -> OrderedDict:
    """
    Convert a string in .mdp format to an OrderedDict

    Values are treated as strings, comments are stripped.

    Args:
        mdp_string (string): Input string in .mdp format (key = value)

    Returns:
        OrderedDict: A dict with the parameters in order
    """

    # add a section heading to satisfy configparser
    mdp_string = "[parameters]\n" + mdp_string
    # provide this string as a file for configparser to read
    mdp_string_as_file = io.StringIO(mdp_string)

    config = configparser.ConfigParser(
        comment_prefixes=[";"], inline_comment_prefixes=[";"])
    config.read_file(mdp_string_as_file)

    return OrderedDict(config['parameters'])



class ParameterNotSetError(KeyError):
    """ Custom exception to handle unset parameters.

        Enable collection of multiple missing parameters to avoid multiple
        runs to uncover missing parameters.
    """
    def __init__(self):
        self.message = ""
        self.have_been_added = False

    def add(self, offending_block, offending_parameter, reason):
        """ add to the collection of key errors

            offending_block : where the error occurred
            message : error message to give
            offending_parameter: the parameter that is not set or set wrongly
        """
        self.message += ("\n" + f"{offending_block} | {offending_parameter}" +
                         f": {reason}")
        self.have_been_added = True


class SomeParametersNotSetError(SystemExit):
    def __init__(self, errors):
        super().__init__("\n".join([error.message for error in errors]))



""" Parse parameters set with --set """
from mdpeditor.mdpblocks import provide
from configparser import ParsingError

def parse_set_parameters(set_value):
    try:
        set_values_as_mdp_string = "\n".join(set_value)
        return provide.mdp_string_to_ordered_dict(set_values_as_mdp_string)
    except ParsingError:
        raise SystemExit("Cannot parse the parameters you are trying to "
                         "set with --set.\nDid you use this format:\t"
                         "--set key=value ?")



""" Run the command line interface """
import argparse
import itertools
import json
import sys

from mdpeditor.mdpblocks import provide
from mdpeditor.mdpblocks.process import apply_instructions
from mdpeditor.parameterhandling import parameters
from mdpeditor.version import __version__

from mdpeditor.commandlineinterface import parsesetparameters


def get_command_line_arguments():
    """ build, parse and return command line arguments """

    program_name = "mdpeditor"
    description = """Compiles an .mdp file from preset .mdp parameter blocks
       and user settings. To learn more about available parameters
       use --describe-blocks.
       """

    epilog = """examples:
    mdpeditor --describe-blocks
    \tShows available pre-defined parameter blocks

    mdpeditor --compile force_field.charmm --set nsteps=100
    \tCompiles the pre-defined block force_field.charmm and
    \tsets "nsteps = 100" in the output

    mdpeditor --compile pressure.atmospheric force_field.charmm --minimal
    \tWrite only values that were set when compiling pressure.atmospheric
    \tforce_field.charmm, skip all defaults
    """

    parser = argparse.ArgumentParser(
        description=description,
        prog=program_name,
        add_help=False,
        epilog=epilog,
        formatter_class=argparse.RawDescriptionHelpFormatter)

    parameters_argument_group = parser.add_argument_group(
        "simulation parameters")

    parameters_argument_group.add_argument(
        "--set",
        nargs="+",
        default="{}",
        help="Set parmaters as parameter1=value1 parameter2=value2")

    parameters_argument_group.add_argument(
        "--compile",
        nargs="+",
        help="pre-defined mdp option blocks to compile an .mdp file, use"
        "--describe-blocks to list all available blocks",
        choices=provide.available_parameter_blocks(),
        metavar="parameter.block")

    parameters_argument_group.add_argument(
        "--input-mdp",
        dest="input_mdp",
        type=argparse.FileType('r'),
        help="read parameters from an .mdp file",
        metavar="input.mdp")

    behaviour_argument_group = parser.add_argument_group("behavior")

    behaviour_argument_group.add_argument(
        "--merge-duplicates",
        dest="merge_right",
        action='store_true',
        default=False,
        help="allow duplicate parameters by overwriting previously set"
        " parameters in the following order, --input-mdp < --compile < --set, "
        " so that parameters specified with --set always prevail")

    parser.add_argument("-h",
                        "--help",
                        action="help",
                        help="show this help message and exit")

    parser.add_argument("--version",
                        action="version",
                        version=f"{program_name} {__version__}")

    inspection_argument_group = parser.add_argument_group("inspect parameters")

    inspection_argument_group.add_argument(
        "--describe-blocks",
        dest="describe_blocks",
        action='store_true',
        default=False,
        help="print description of all available parameter blocks and exit")

    inspection_argument_group.add_argument(
        "--show-block",
        dest="show_block",
        choices=provide.available_parameter_blocks(),
        help="show the selected parameter block and exit",
        metavar="block")

    output_argument_group = parser.add_argument_group("output")

    output_argument_group.add_argument(
        "--output",
        nargs='?',
        const='compiled.mdp',
        type=argparse.FileType('w'),
        help="write the compiled parameters to an .mdp file"
        " (instead of command line)",
        metavar="compiled.mdp")

    output_argument_group.add_argument("--minimal",
                                       dest="minimal",
                                       action='store_true',
                                       default=False,
                                       help="keep the output minimal")

    # print the help text if no arguments are given
    if len(sys.argv) == 1:
        return parser.parse_args(args=["--help"])

    return parser.parse_args()


def run():
    """ run the command line interface """
    command_line_arguments = get_command_line_arguments()

    if (command_line_arguments.describe_blocks
            and command_line_arguments.show_block):
        raise SystemExit(
            "Cannot --show-block and --describe-blocks at the same time.")

    if command_line_arguments.describe_blocks:
        print("\nPre-defined parameter blocks for --compile\n")
        print(provide.describe_parameter_blocks())
        sys.exit()

    if command_line_arguments.show_block:
        print(
            provide.parameter_block_as_string(
                command_line_arguments.show_block))
        sys.exit()

    # read user parameter file
    mdp_input_as_string = ""
    if command_line_arguments.input_mdp:
        mdp_input_as_string = command_line_arguments.input_mdp.read()
    mdp_input = provide.mdp_string_to_ordered_dict(mdp_input_as_string)

    # read parameter blocks
    mdp_option_blocks = provide.read_parameter_blocks(
        command_line_arguments.compile)

    # add user-provided parameters
    input_parameter_blocks = itertools.chain(
        [mdp_input], mdp_option_blocks,
        [parsesetparameters.parse_set_parameters(
            command_line_arguments.set)])

    compiled_parameters = parameters.compile_parameters(input_parameter_blocks)

    if (not command_line_arguments.merge_right
            and compiled_parameters.duplicate_keys):
        raise SystemExit(
            "\nAborting compilation due to duplicate parameter(s)\n\n\t" +
            "\n\t".join(list(compiled_parameters.duplicate_keys)) +
            "\n\nUse --merge-duplicates to override parameters\n")

    apply_instructions(compiled_parameters.parameters,
                       command_line_arguments.compile)

    output_parameters = parameters.OutputParameters(
        compiled_parameters.parameters)

    if command_line_arguments.minimal:
        # discard all parameters that were not explicitely chosen
        output_parameters.keep_only(compiled_parameters.parameters.keys())

    writing_to_terminal = True if (not command_line_arguments.output
                                   and sys.stdout.isatty()) else False
    # color output only if we have an interactive terminal and
    # a full-blown .mdp is asked for
    if writing_to_terminal and not command_line_arguments.minimal:
        output_string = output_parameters.as_colored_string(
            compiled_parameters.parameters.keys())
    else:
        output_string = output_parameters.as_string()

    # if we write or pipe to a file, keep track of the command used to generate
    # the output by prepending a commented line
    if not writing_to_terminal:
        output_string = f"; Created by mdpeditor version {__version__}\n" + \
                        "; " + ' '.join(sys.argv) + '\n' + output_string

    # write either to terminal or to file
    if command_line_arguments.output:
        command_line_arguments.output.write(output_string)
    else:
        print(output_string)



from mdpeditor.mdpblocks.parameternotseterror import ParameterNotSetError


def not_set_or_empty(dictionary, key):
    """ Returns true if a key is not set or empty """
    return (key not in dictionary.keys() or not dictionary[key].strip())


def process(parameters, block_name):
    """ Set density guided simulation parameters from processing input

        Consume parameters that are not mdp options, but rather used to
        calculate these.

        Sets

            density-guided-simulation-gaussian-transform-spreading-width
            density-guided-simulation-reference-density-filename

        parameters : the full set of parameters, including non-mdp options
        block_name : allows error reporting.
    """
    pixel_size_key = "pixel-size-in-nm"

    errors = ParameterNotSetError()

    # set spreading width from pixel size if its not defined or empty
    width_key = "density-guided-simulation-gaussian-transform-spreading-width"
    if not_set_or_empty(parameters, width_key):
        try:
            parameters[width_key] = str(0.85 *
                                        float(parameters[pixel_size_key]))
            del parameters[pixel_size_key]
        except ValueError:
            errors.add(
                block_name, width_key,
                f" cannot convert \"{pixel_size_key}\"" +
                f"={parameters[pixel_size_key]} "
                "to a spreading width")
        except KeyError:
            errors.add(
                block_name, width_key, " not set in input, use --set " +
                f"{pixel_size_key}=YOURVALUE")
    else:
        if pixel_size_key in parameters:
            errors.add(
                block_name, width_key,
                "you have set a pixel size to determine the spreading width," +
                f" but {width_key} is already set to " +
                f"{parameters[width_key]}")

    # handle the reference density file name
    reference_file_key = "density-guided-simulation-reference-density-filename"
    reference_file_key_short = "reference-density-filename"

    if not_set_or_empty(parameters, reference_file_key):
        try:
            parameters[reference_file_key] = parameters[
                reference_file_key_short]
        except KeyError:
            errors.add(
                block_name, reference_file_key,
                "the reference density file name is not set" +
                ", use the shorthand --set " +
                f"{reference_file_key_short}=FILENAME")

    if errors.have_been_added:
        raise errors



""" Perform parameter calculations.

    Add new processing instructions for parameter blocks by adding to
    the processing_instructions dict.

    Note: This has to be kept in sync with how parameter block naming and
    discovery works in the provide package in this sub-module; ideally they
    would both share the discovery code.
"""
from mdpeditor.mdpblocks.parameternotseterror import (ParameterNotSetError,
                                                      SomeParametersNotSetError)
import mdpeditor.mdpblocks.density_guided.vanilla


def apply_instructions(parameters, block_names: str):
    """ Perform instructions by the parameter blocks
        Blocks may fill in a parameters that are needed by other blocks.

        Run through all blocks two times, being forgiving if
        keys are not found the first time, because they might be filled in
        by another block later.
    """
    processing_instructions = {
        'density_guided.vanilla':
        mdpeditor.mdpblocks.density_guided.vanilla.process,
    }

    # keep only the block names that have processing instructions
    block_names_to_process = [
        name for name in block_names if name in processing_instructions
    ]

    # try to evaluate all parameters by block-defined procedures, but don't do
    # anything if a key is not yet defined
    for block_name in block_names_to_process:
        try:
            processing_instructions[block_name](parameters, block_name)
        except ParameterNotSetError:
            pass

    # during the second try, undefined parameters are an issue
    #   we collect errors across all blocks, report and then abort.
    errors = []
    for block_name in block_names_to_process:
        try:
            processing_instructions[block_name](block_name=block_name,
                                                parameters=parameters)
        except ParameterNotSetError as e:
            print(block_name)
            errors.append(e)

    if errors:
        raise SomeParametersNotSetError(errors)
